// SPDX-FileCopyrightText: 2026 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#pragma once

namespace linyaps_box {

struct oci_config;

void validate(const oci_config &config);

} // namespace linyaps_box
