// SPDX-FileCopyrightText: 2025 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#include <linyaps_box/cgroup_manager.h>

namespace linyaps_box {

cgroup_manager::~cgroup_manager() = default;

} // namespace linyaps_box
