// SPDX-FileCopyrightText: 2025 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#include "linyaps_box/interface.h"

namespace linyaps_box {

// ensure vtable only here
interface::~interface() = default;

} // namespace linyaps_box
