// SPDX-FileCopyrightText: 2022-2025 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#pragma once

namespace linyaps_box {

auto main(int argc, char **argv) noexcept -> int;

} // namespace linyaps_box
